# -*- coding: utf-8 -*-

import plotly.graph_objs as go
import numpy as np
import seaborn as sns
import thunor.config as config
from .helpers import format_dose
from .dip import ctrl_dip_rates, expt_dip_rates
from thunor.curve_fit import HillCurveNull, is_param_truncated
import scipy.stats
import re
import pandas as pd
import collections


class CannotPlotError(ValueError):
    pass


def _activity_area_units(**kwargs):
    if 'aa_max_conc' in kwargs:
        return '[max. dose={}]'.format(format_dose(kwargs['aa_max_conc']))
    else:
        return ''


def _auc_units(**kwargs):
    if 'auc_min_conc' in kwargs:
        return '[min. dose={}]'.format(format_dose(kwargs['auc_min_conc']))
    else:
        return ''


SECONDS_IN_HOUR = 3600.0
NS_IN_SEC = 1e9
PLATE_MAP_WELL_DIAM = 0.95
ASCII_CAP_A = 65
PARAM_UNITS = {'auc': _activity_area_units,
               'aa': _activity_area_units,
               'einf': 'h<sup>-1</sup>',
               'emax': 'h<sup>-1</sup>',
               'emax_obs': 'h<sup>-1</sup>'}
PARAM_NAMES = {'aa': 'Activity area',
               'aa_obs': 'Activity area (observed)',
               'aa_num': 'Activity area (numerical integration)',
               'auc': 'Area under curve',
               'einf': 'E<sub>inf</sub>',
               'emax': 'E<sub>max</sub>',
               'emax_rel': 'E<sub>max</sub> (relative)',
               'emax_obs': 'E<sub>max</sub> observed',
               'emax_obs_rel': 'E<sub>Max</sub> observed (relative)',
               'hill': 'Hill coefficient'}
IC_REGEX = re.compile('^ic([0-9]+)$')
EC_REGEX = re.compile('^ec([0-9]+)$')
E_REGEX = re.compile('^e([0-9]+)$')
E_REL_REGEX = re.compile('^e([0-9]+)_rel$')


def _secs_to_str(seconds):
    """
    Convert seconds to HH:MM:SS string

    Parameters
    ----------
    seconds: int
        Number of seconds

    Returns
    -------
    str
        Formatted string like "Time: 2:04:00"
    """
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return "Time: %d:%02d:%02d" % (h, m, s)


def _remove_drmetric_prefix(param_id):
    """ Remove DR metric prefix on parameter name, if present """
    for prefix in ('viability__', 'dip__'):
        if param_id.startswith(prefix):
            param_id = param_id[len(prefix):]

    return param_id


def _param_is_log(param_id):
    param_id = _remove_drmetric_prefix(param_id)
    return (IC_REGEX.match(param_id) or EC_REGEX.match(param_id) or
            param_id == 'hill')


def _param_na_first(param_id):
    """ NAs go in the position corresponding to no effect """
    param_id = _remove_drmetric_prefix(param_id)
    # Which is first for E, Emax, Erel, AA and Hill
    return param_id in ('hill', 'aa', 'emax', 'emax_rel', 'einf') \
           or E_REGEX.match(param_id) \
           or E_REL_REGEX.match(param_id)


def _get_param_name(param_id):
    param_id = _remove_drmetric_prefix(param_id)
    try:
        return PARAM_NAMES[param_id]
    except KeyError:
        if IC_REGEX.match(param_id):
            return 'IC<sub>{:d}</sub>'.format(int(param_id[2:]))
        elif EC_REGEX.match(param_id):
            return 'EC<sub>{:d}</sub>'.format(int(param_id[2:]))
        elif E_REGEX.match(param_id):
            return 'E<sub>{:d}</sub>'.format(int(param_id[1:]))
        elif E_REL_REGEX.match(param_id):
            return 'E<sub>{:d}</sub> (relative)'.format(
                int(param_id[1:param_id.index('_')]))
        else:
            return param_id


def _get_param_units(param_id):
    param_id = _remove_drmetric_prefix(param_id)
    try:
        return PARAM_UNITS[param_id]
    except KeyError:
        if IC_REGEX.match(param_id) or EC_REGEX.match(param_id):
            return 'M'
        elif E_REGEX.match(param_id):
            return 'h<sup>-1</sup>'
        else:
            return ''


def _out_of_range_msg(param_id):
    return '{} truncated to limit of measured concentrations'.format(
        _get_param_name(param_id)
    )


def _sns_to_rgb(palette):
    return ['rgb(%d, %d, %d)' % (c[0] * 255, c[1] * 255, c[2] * 255)
            if not isinstance(c, str) else c
            for c in palette]


def _make_title(title, df):
    drug_list = df.index.get_level_values('drug').unique()
    if len(drug_list) == 1:
        drug_name = drug_list[0]
        if not isinstance(drug_name, str):
            if isinstance(drug_name, collections.Iterable):
                if len(drug_name) == 1:
                    drug_name = drug_name[0]
                else:
                    drug_name = " &amp; ".join(drug_name)
            else:
                raise ValueError('Unknown drug_name type: {}'.format(type(
                    drug_name)))
        title += ' for {}'.format(drug_name)

    cell_line_list = df.index.get_level_values('cell_line').unique()
    if len(cell_line_list) == 1:
        title += ' on {}'.format(cell_line_list[0])

    return title


def _combine_title_subtitle(title, subtitle):
    if subtitle:
        title += '<br> <span style="color:#999;font-size:0.9em">' \
                 '{}</span>'.format(subtitle)

    return title


def plot_drc(fit_params, is_absolute=False, color_by=None, color_groups=None,
             title=None, subtitle=None, template=config.plotly_template):
    """
    Plot dose response curve fits

    Parameters
    ----------
    fit_params: pd.DataFrame
        Fit parameters from :func:`thunor.curve_fit.fit_params`
    is_absolute: bool
        For DIP rate plots, use absolute (True) or relative (False)
        y-axis scale. **Ignored for viability plots.**
    color_by: str or None
        Color the traces by cell lines if 'cl', drugs if 'dr',
        or arbitrarily if None (default)
    color_groups: dict or None
        If using color_by, provide a dictionary containing the color groups,
        where the values are cell line or drug names
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    template: str
        Name of plotly template (https://plot.ly/python/templates/)

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    datasets = fit_params.index.get_level_values('dataset_id').unique()
    if color_by:
        num_colors = len(color_groups)
    elif len(datasets) == 2:
        num_colors = 2
    else:
        num_colors = len(fit_params)

    colours = _sns_to_rgb(sns.color_palette("husl", num_colors))
    color_index_col = None
    if color_by == 'cl':
        color_index_col = fit_params.index.names.index('cell_line')
    elif color_by == 'dr':
        color_index_col = fit_params.index.names.index('drug')
    elif color_by is None:
        if len(datasets) == 2:
            color_by = 'dataset'
            color_index_col = fit_params.index.names.index('dataset_id')
            color_groups = {dataset: [dataset] for dataset in datasets}
    else:
        raise ValueError('color_by must be "cl", "dr" or None')
    # Shapes used for replicate markers
    shapes = ['circle', 'circle-open']

    try:
        is_viability = 'viability' in fit_params.columns or \
                       fit_params._drmetric == 'viability'
    except AttributeError:
        is_viability = False

    if is_viability:
        # Only "absolute" (non-transformed y-axis) makes sense for viability
        is_absolute = True

    if is_viability:
        yaxis_title = '{:g} hr viability'.format(fit_params._viability_time)
    else:
        yaxis_title = 'DIP rate'
    if is_absolute:
        if not is_viability:
            yaxis_title += ' (h<sup>-1</sup>)'
    else:
        yaxis_title = 'Relative ' + yaxis_title

    multi_dataset = len(datasets) > 1
    show_annotations = len(fit_params.index) == 1

    show_replicates = len(fit_params.index) == 1 or \
        (multi_dataset and
         len(fit_params.index.get_level_values('cell_line').unique()) == 1 and
         len(fit_params.index.get_level_values('drug').unique()) == 1)

    if title is None:
        title = _make_title('Dose response', fit_params)
    if subtitle is None:
        subtitle = " &amp; ".join(str(d) for d in datasets)
    title = _combine_title_subtitle(title, subtitle)

    if show_annotations:
        fit_params = fit_params.copy()
        fit_params['ec50_truncated'] = is_param_truncated(fit_params, 'ec50')
        fit_params['ic50_truncated'] = is_param_truncated(fit_params, 'ic50')

    annotations = []
    traces = []
    if color_by:
        for idx, name in enumerate(color_groups):
            traces.append(go.Scatter(mode='none', x=[0], y=[0],
                                     legendgroup=name,
                                     showlegend=True,
                                     name='<b>{}</b>'.format(name))
                          )
    xaxis_min = np.Inf
    xaxis_max = np.NINF
    for fp in fit_params.itertuples():
        if color_by:
            grp = fp.Index[color_index_col]
            this_colour = None
            legend_grp = None
            for idx, color_group in enumerate(color_groups.items()):
                grp_label, grp_entries = color_group
                # Not an efficient lookup, but for reasonable number of
                # traces it's fine
                if grp in grp_entries:
                    legend_grp = grp_label
                    this_colour = colours[idx]
                    break
            if this_colour is None:
                raise ValueError('"{}" is not in the color_groups'.format(grp))
        else:
            this_colour = colours.pop()
            legend_grp = fp.label

        group_name_disp = fp.label

        log_dose_min = int(np.floor(np.log10(fp.min_dose_measured)))
        log_dose_max = int(np.ceil(np.log10(fp.max_dose_measured)))

        try:
            if is_viability:
                ctrl_doses = fp.viability_ctrl.index.get_level_values('dose')
            else:
                ctrl_doses = fp.dip_ctrl.index.get_level_values('dose')

            log_dose_min = min(log_dose_min, int(np.floor(np.log10(np.min(
                ctrl_doses.values)))))
        except AttributeError:
            ctrl_doses = []

        xaxis_max = max(xaxis_max, log_dose_max)
        xaxis_min = min(xaxis_min, log_dose_min)

        try:
            if is_viability:
                expt_doses = fp.viability.index.get_level_values('dose')
            else:
                expt_doses = fp.dip_expt.index.get_level_values('dose')
        except AttributeError:
            expt_doses = []

        dose_x_range = np.concatenate(
            # [np.arange(2, 11) * 10 ** dose_mag
            [0.5 * np.arange(3, 21) * 10 ** dose_mag
             for dose_mag in range(log_dose_min, log_dose_max + 1)],
            axis=0)

        dose_x_range = np.append([10 ** log_dose_min], dose_x_range,
                                 axis=0)

        line_dash = 'solid'
        line_mode = 'lines'
        hoverinfo = 'all'
        visible = True
        if fp.fit_obj is None:
            # Curve fit numerical error or QC failure
            dose_x_range = None
            dip_rate_fit = None
            line_mode = 'none'
            group_name_disp = '<i>{}</i>'.format(
                group_name_disp)
            hoverinfo = 'none'
            visible = 'legendonly'
        elif isinstance(fp.fit_obj, HillCurveNull):
            # No effect null hypothesis
            dip_rate_fit = [1 if not is_absolute else fp.fit_obj.divisor] * \
                            len(dose_x_range)
            line_dash = 'longdash'
        else:
            # Fit succeeded
            if is_absolute:
                dip_rate_fit = fp.fit_obj.fit(dose_x_range)
            else:
                dip_rate_fit = fp.fit_obj.fit_rel(dose_x_range)

        traces.append(go.Scatter(x=dose_x_range,
                                 y=dip_rate_fit,
                                 mode=line_mode,
                                 line={'shape': 'spline',
                                       'color': this_colour,
                                       'dash': line_dash,
                                       'width': 3},
                                 hoverinfo=hoverinfo,
                                 legendgroup=legend_grp,
                                 showlegend=not show_replicates or
                                            multi_dataset,
                                 visible=visible,
                                 name=group_name_disp)
                      )

        if show_replicates:
            y_trace = fp.viability if is_viability else fp.dip_expt
            try:
                ctrl_resp = fp.viability_ctrl if is_viability else fp.dip_ctrl
            except AttributeError:
                ctrl_resp = None
            if not is_absolute:
                if fp.fit_obj is not None:
                    divisor = fp.fit_obj.divisor
                elif fp.fit_obj is None and ctrl_resp is not None:
                    divisor = np.mean(ctrl_resp)
                else:
                    divisor = 1
                y_trace /= divisor
                if ctrl_resp is not None:
                    ctrl_resp /= divisor

            repl_name = 'Replicate'
            ctrl_name = 'Control'
            if multi_dataset:
                repl_name = '{} {}'.format(fp.Index[0], repl_name)
                ctrl_name = '{} {}'.format(fp.Index[0], ctrl_name)

            if is_viability:
                # viability times are in nanoseconds - convert
                hoverlabels = [_secs_to_str(int(x) / 1e9) for x in
                               fp.viability_time]
            else:
                hoverlabels = repl_name

            shape = shapes.pop(0)

            traces.append(go.Scatter(x=expt_doses,
                                     y=y_trace,
                                     mode='markers',
                                     marker={'symbol': shape,
                                             'color': this_colour,
                                             'size': 5},
                                     legendgroup=group_name_disp,
                                     hoverinfo='x+y+text',
                                     text=hoverlabels,
                                     showlegend=False,
                                     name=repl_name)
                          )
            if ctrl_resp is not None:
                ctrl_color = 'white' if template == 'plotly_dark' else 'black'
                traces.append(go.Scatter(x=ctrl_doses,
                                         y=ctrl_resp,
                                         mode='markers',
                                         marker={'symbol': shape,
                                                 'color': ctrl_color,
                                                 'size': 5},
                                         hoverinfo='y+text',
                                         text=ctrl_name,
                                         name=ctrl_name,
                                         legendgroup=group_name_disp,
                                         showlegend=False)
                              )

        if show_annotations:
            annotation_label = ''
            if fp.ec50 is not None:
                annotation_label += 'EC<sub>50</sub>{}: {} '.format(
                    '*' if fp.ec50_truncated else '',
                    format_dose(fp.ec50, sig_digits=5)
                )
            if fp.ic50 is not None:
                annotation_label += 'IC<sub>50</sub>{}: {} '.format(
                    '*' if fp.ic50_truncated else '',
                    format_dose(fp.ic50, sig_digits=5)
                )
            if fp.emax is not None:
                annotation_label += 'E<sub>max{}</sub>{}: {:.5g}'.format(
                    ' rel' if not is_absolute else '',
                    '*' if fp.fit_obj.emax < fp.emax else '',
                    fp.emax if is_absolute else fp.emax_rel)
            if annotation_label:
                hovermsgs = []
                hovertext = None
                if fp.ec50_truncated:
                    hovermsgs.append(_out_of_range_msg('ec50'))
                if fp.ic50_truncated:
                    hovermsgs.append(_out_of_range_msg('ic50'))
                if hovermsgs:
                    hovertext = '*' + '<br>'.join(hovermsgs)
                annotations.append({
                    'x': 0.5,
                    'y': 1.0,
                    'xref': 'paper',
                    'yanchor': 'bottom',
                    'yref': 'paper',
                    'showarrow': False,
                    'hovertext': hovertext,
                    'text': annotation_label
                })

    yaxis_range = None
    yaxis_rangemode = 'tozero' if is_viability else 'normal'
    if not is_absolute:
        yaxis_range = (-0.2, 1.2)
    elif not is_viability:
        yaxis_range = (-0.02, 0.07)

    layout = go.Layout(title=title,
                       hovermode='closest' if show_replicates
                                 or len(traces) > 50 else 'x',
                       xaxis={'title': 'Dose (M)',
                              'range': (xaxis_min, xaxis_max),
                              'type': 'log'},
                       yaxis={'title': yaxis_title,
                              'range': yaxis_range,
                              'rangemode': yaxis_rangemode
                              },
                       annotations=annotations,
                       template=template
                       )

    return go.Figure(data=traces, layout=layout)


def plot_drug_combination_heatmap(
        ctrl_resp_data, expt_resp_data, title=None, subtitle=None,
        template=config.plotly_template):
    """
    Plot heatmap of drug combination response by DIP rate

    Two dimensional plot (each dimension is a drug concentration) where
    squares are coloured by DIP rate value.

    Parameters
    ----------
    ctrl_resp_data: pd.DataFrame
        Control DIP rates from :func:`thunor.dip.dip_rates`
    expt_resp_data: pd.DataFrame
        Experiment (non-control) DIP rates from :func:`thunor.dip.dip_rates`
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    template: str
        Name of plotly template (https://plot.ly/python/templates/)

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    heat_label = 'DIP<br>rate'

    if title is None:
        title = _make_title('Dose response', expt_resp_data)

    if subtitle is None:
        datasets = expt_resp_data.index.get_level_values('dataset').unique()
        subtitle = datasets[0]

    title = _combine_title_subtitle(title, subtitle)

    expt_resp_data = expt_resp_data['dip_rate']
    if ctrl_resp_data is None:
        raise CannotPlotError('There are no matching control wells for this '
                              'selection, so relative DIP rate cannot be '
                              'calculated')

    expt_resp_data = expt_resp_data / ctrl_resp_data['dip_rate'].mean()
    heat_label = 'Relative<br>' + heat_label

    expt_resp_data = expt_resp_data.reset_index(['dose', 'drug'])
    doses = expt_resp_data['dose'].apply(pd.Series)
    doses.columns = ['dose1', 'dose2']

    drugs = expt_resp_data['drug'].apply(pd.Series)
    drugs.columns = ['drug1', 'drug2']
    drug1 = drugs['drug1'].unique()[0]
    drug2 = drugs['drug2'].unique()[0]

    expt_resp_data = pd.concat([doses, drugs,
                                expt_resp_data['dip_rate']], axis=1)

    expt_resp_data = expt_resp_data.set_index(['dose1', 'dose2'])

    dat = []
    dose1 = sorted(expt_resp_data.index.get_level_values('dose1').unique())
    dose2 = sorted(expt_resp_data.index.get_level_values('dose2').unique())
    for d1, grp in expt_resp_data['dip_rate'].groupby('dose1'):
        dat2 = []
        grp.index = grp.index.droplevel()
        for d2 in dose2:
            try:
                dat2.append(grp.loc[d2].mean())
            except KeyError:
                if d1 == 0 and d2 == 0:
                    # Control well relative DIP rate is 1.0 by definition
                    dat2.append(1.0)
                else:
                    dat2.append(None)
        dat.append(dat2)

    trace = go.Heatmap(x=format_dose(dose2), y=format_dose(dose1), z=dat,
                       colorbar={'title': heat_label}, zmin=-1.5, zmax=1.5,
                       colorscale=[
                           (0, 'rgb(255,0,0)'),
                           (0.5, 'rgb(255,255,0)'),
                           (1, 'rgb(0,0,255)')
                       ])

    layout = go.Layout(
        title=title,
        xaxis={
            'title': '{} concentration'.format(drug2)
        },
        yaxis={
            'title': '{} concentration'.format(drug1)
        },
        template=template
    )

    return go.Figure(data=[trace], layout=layout)


def _symbols_hovertext_two_dataset_scatter(df_params, range_bounded_params,
                                           fit_param, dataset_names):
    symbols = ['circle'] * len(df_params.index)
    hovertext = [" ".join(l) for l in df_params.index.values]
    for param in range_bounded_params:
        msg = _out_of_range_msg(param)
        for i in (0, 1):
            tmp_df = pd.concat([
                df_params.loc[:, fit_param].iloc[:, i],
                df_params.loc[:, 'max_dose_measured'].iloc[:, i],
                df_params.loc[:, 'min_dose_measured'].iloc[:, i]
            ], axis=1)
            tmp_df.columns = [fit_param,
                              'max_dose_measured',
                              'min_dose_measured']
            param_truncated = is_param_truncated(tmp_df, fit_param)
            addtxt = ['<br> {} {}'.format(dataset_names[i], msg) if x else
                      '' for x in param_truncated]
            hovertext = [ht + at for ht, at in zip(hovertext, addtxt)]
            symbols = ['cross' if x else old for x, old in
                       zip(param_truncated, symbols)]

    return symbols, hovertext


def plot_two_dataset_param_scatter(df_params, fit_param, title, subtitle,
                                   color_by, color_groups,
                                   template=config.plotly_template, **kwargs):
    """
    Plot a parameter comparison across two datasets

    Parameters
    ----------
    df_params: pd.DataFrame
        DIP fit parameters from :func:`thunor.dip.dip_params`
    fit_param: str
        The name of the parameter to compare across datasets, e.g. ic50
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    template: str
        Name of plotly template (https://plot.ly/python/templates/)
    kwargs: dict, optional
        Additional keyword arguments

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    if title is None:
        title = _make_title('Dose response parameters', df_params)
    if subtitle is None:
        datasets = df_params.index.get_level_values('dataset_id').unique()
        subtitle = " &amp; ".join(str(d) for d in datasets)
    title = _combine_title_subtitle(title, subtitle)

    if df_params._drmetric == 'dip':
        dr_metric = 'DIP'
    else:
        dr_metric = '{:g} hr viability'.format(df_params._viability_time)

    df_params = df_params.loc[:, [fit_param,
                                  'max_dose_measured',
                                  'min_dose_measured']]
    df_params.dropna(subset=[fit_param], inplace=True)
    df_params.reset_index(inplace=True)
    df_params = df_params.pivot_table(index=['cell_line', 'drug'], columns=[
        'dataset_id'])
    df_params.dropna(inplace=True)

    if len(df_params.index) == 0:
        raise CannotPlotError(
            'Dataset vs dataset scatter plot is empty. Check the cell lines '
            'and drugs in the two datasets overlap. If you want a bar plot '
            'instead, choose an ordering parameter.')

    if len(df_params.columns) < 4:
        raise CannotPlotError(
            'The cell lines and/or drugs selected only have data in one of '
            'the two datasets, so a scatter plot cannot be created. If you '
            'want a bar plot instead, choose an ordering parameter.')

    if color_by:
        colours = _sns_to_rgb(sns.color_palette("husl", len(color_groups)))
    else:
        colours = _sns_to_rgb(sns.color_palette("Paired"))[0:2]

    param_name = _get_param_name(fit_param)
    param_units = _get_param_units(fit_param)
    try:
        param_units = param_units(**kwargs)
    except TypeError:
        pass
    if param_units:
        axis_title = '{} ({})'.format(param_name, param_units)
    else:
        axis_title = param_name

    axis_title = '{} {}'.format(dr_metric, axis_title)

    fit_param_data = df_params.loc[:, fit_param]

    range_bounded_params = set()

    match = IC_REGEX.match(fit_param)
    if match:
        range_bounded_params.add(match.group())
    match = EC_REGEX.match(fit_param)
    if match:
        range_bounded_params.add(match.group())
    match = E_REGEX.match(fit_param)
    if match:
        range_bounded_params.add('ec' + match.groups(0)[0])
    match = E_REL_REGEX.match(fit_param)
    if match:
        range_bounded_params.add('ec' + match.groups(0)[0])

    dataset_names = fit_param_data.columns

    symbols, hovertext = _symbols_hovertext_two_dataset_scatter(
        df_params, range_bounded_params, fit_param, dataset_names
    )

    xdat = fit_param_data.iloc[:, 0]
    xdat_fit = xdat
    ydat = fit_param_data.iloc[:, 1]
    ydat_fit = ydat
    if _param_is_log(fit_param):
        xdat_fit = np.log10(xdat)
        ydat_fit = np.log10(ydat)

    fitdat_mask = (~np.isnan(xdat_fit) & np.isfinite(xdat_fit) &
                   ~np.isnan(ydat_fit) & np.isfinite(ydat_fit) &
                   [s != 'cross' for s in symbols])
    xdat_fit = xdat_fit[fitdat_mask].values
    ydat_fit = ydat_fit[fitdat_mask].values

    data = []
    layout = go.Layout(title=title)

    slope, intercept, r_value, p_value, std_err = \
        scipy.stats.linregress(xdat_fit, ydat_fit)
    if not np.isnan(slope):
        xfit = (min(xdat_fit), max(xdat_fit))
        yfit = [x * slope + intercept for x in xfit]
        if _param_is_log(fit_param):
            xfit = np.power(10, xfit)
            yfit = np.power(10, yfit)
        data.append(go.Scatter(
            x=xfit,
            y=yfit,
            mode='lines',
            hoverinfo="none",
            line=dict(
                color="darkorange"
            ),
            name='{} vs {} {} Linear Fit'.format(dataset_names[0],
                                      dataset_names[1],
                                      param_name),
            showlegend=False
        ))
        layout['annotations'] = [{
            'x': 0.5, 'y': 1.0, 'xref': 'paper', 'yanchor': 'bottom',
            'yref': 'paper', 'showarrow': False,
            'text': 'R<sup>2</sup>: {:0.4g} '
                    'p-value: {:0.4g} '.format(r_value ** 2, p_value)
        }]

    custom_data = [
        {'c': cl, 'd': dr} for cl, dr in zip(
            df_params.index.get_level_values('cell_line'),
            df_params.index.get_level_values('drug')
        )
    ]

    if color_by:
        for idx, tag_name in enumerate(color_groups):
            dat = df_params[df_params.index.get_level_values(
                'cell_line' if color_by == 'cl' else 'drug').isin(
                color_groups[tag_name])]
            symbols, hovertext = _symbols_hovertext_two_dataset_scatter(dat, range_bounded_params,
                                                                        fit_param, dataset_names)

            fit_param_data = dat.loc[:, fit_param]
            xdat = fit_param_data.iloc[:, 0]
            ydat = fit_param_data.iloc[:, 1]

            data.append(go.Scatter(
                x=xdat,
                y=ydat,
                hovertext=hovertext,
                hoverinfo="text+x+y",
                mode='markers',
                customdata=custom_data,
                marker={'symbol': symbols,
                        'color': colours[idx]},
                name=tag_name
            ))
    else:
        colour_list = [colours[1] if s == 'circle' else 'crimson' for s in
                       symbols]

        data.append(go.Scatter(
            x=xdat,
            y=ydat,
            hovertext=hovertext,
            hoverinfo="text+x+y",
            mode='markers',
            customdata=custom_data,
            marker={'symbol': symbols,
                    'color': colour_list},
            name='{} vs {} {}'.format(dataset_names[0],
                                      dataset_names[1],
                                      param_name)
        ))

    layout['xaxis'] = {'title': '{} {}'.format(dataset_names[0], axis_title),
                       'type': 'log' if _param_is_log(fit_param) else None}
    layout['yaxis'] = {'title': '{} {}'.format(dataset_names[1], axis_title),
                       'type': 'log' if _param_is_log(fit_param) else None}
    layout['hovermode'] = 'closest'
    layout['showlegend'] = color_by is not None
    layout['template'] = template

    return go.Figure(layout=layout, data=data)


def _symbols_hovertext_two_param_scatter(df_params,
                                         range_bounded_params):
    hovertext = df_params['label']
    symbols = ['circle'] * len(df_params)
    for param in range_bounded_params:
        msg = _out_of_range_msg(param)
        param_truncated = is_param_truncated(df_params, param)
        addtxt = ['<br> ' + msg if x else '' for x in
                  param_truncated]
        hovertext = [ht + at for ht, at in zip(hovertext, addtxt)]
        symbols = ['cross' if x else old for x, old in
                   zip(param_truncated, symbols)]
    return symbols, hovertext


def plot_drc_params(df_params, fit_param,
                    fit_param_compare=None,
                    fit_param_sort=None,
                    title=None,
                    subtitle=None,
                    aggregate_cell_lines=False,
                    aggregate_drugs=False,
                    multi_dataset=False,
                    color_by=None,
                    color_groups=None,
                    template=config.plotly_template,
                    **kwargs):
    """
    Box, bar, or scatter plots of DIP rate fit parameters

    Parameters
    ----------
    df_params: pd.DataFrame
        DIP fit parameters from :func:`thunor.dip.dip_params`
    fit_param: str
        Fit parameter name, e.g. 'ic50'
    fit_param_compare: str, optional
        Second fit parameter name for comparative plots, e.g. 'ec50'
    fit_param_sort: str, optional
        Fit parameter name to use for sorting the x-axis, if different from
        fit_param
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    aggregate_cell_lines: bool or dict, optional
        Aggregate all cell lines (if True), or aggregate by the specified
        groups (dict of cell line names as values, with group labels as keys)
    aggregate_drugs: bool or dict, optional
        Aggregate all drugs (if True), or aggregate by the specified
        groups (dict of drug names as values, with group labels as keys)
    multi_dataset: bool
        Set to true to compare two datasets contained in fit_params
    color_by: str or None
        Color by cell lines if "cl", drugs if "dr", or arbitrarily if None
        (default)
    color_groups: dict or None
        Groups of cell lines of drugs to color by
    template: str
        Name of plotly template (https://plot.ly/python/templates/)
    kwargs: dict, optional
        Additional keyword arguments

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    if fit_param_compare and (aggregate_cell_lines or aggregate_drugs):
        raise CannotPlotError(
            'Aggregation is not available when comparing two dose response '
            'parameters')

    if multi_dataset and fit_param_compare is None and \
            not aggregate_cell_lines and \
            not aggregate_drugs and fit_param_sort is None:
        return plot_two_dataset_param_scatter(
            df_params,
            fit_param,
            title,
            subtitle,
            color_by,
            color_groups,
            **kwargs
        )

    color_by_col = None
    if multi_dataset and not color_by:
        color_by_col = 'dataset_id'
        color_by = 'dataset'
        color_groups = {dataset: [dataset] for dataset in df_params.index.get_level_values('dataset_id').unique()}
        colours = _sns_to_rgb(sns.color_palette("husl", 2))
    elif color_by:
        color_by_col = 'cell_line' if color_by == 'cl' else 'drug'
        colours = _sns_to_rgb(sns.color_palette("husl", len(color_groups)))
    else:
        colours = _sns_to_rgb(sns.color_palette("Paired"))[0:2]

    if title is None:
        title = _make_title('Dose response parameters', df_params)
    if subtitle is None:
        datasets = df_params.index.get_level_values('dataset_id').unique()
        subtitle = " &amp; ".join(str(d) for d in datasets)
    title = _combine_title_subtitle(title, subtitle)

    yaxis_param_name = _get_param_name(fit_param)
    yaxis_units = _get_param_units(fit_param)
    try:
        yaxis_units = yaxis_units(**kwargs)
    except TypeError:
        pass
    if yaxis_units:
        yaxis_title = '{} ({})'.format(yaxis_param_name, yaxis_units)
    else:
        yaxis_title = yaxis_param_name
    if df_params._drmetric in ('dip', 'compare'):
        yaxis_title = 'DIP {}'.format(yaxis_title)
    else:
        yaxis_title = '{:g} hr viability {}'.format(df_params._viability_time,
                                                     yaxis_title)

    layout = dict(title=title,
                  template=template,
                  yaxis={'title': yaxis_title,
                         'type': 'log' if _param_is_log(fit_param) else None})

    if fit_param_compare:
        df_params.dropna(subset=[fit_param, fit_param_compare],
                         inplace=True)
        if df_params.empty:
            raise CannotPlotError(
                'No data exists for this selection. This may be due to '
                'missing drug/cell line combinations, or undefined parameters '
                'for the selection.')
        if fit_param == fit_param_compare:
            xdat = df_params.loc[:, fit_param]
            ydat = xdat
        else:
            dat = df_params.loc[:, [fit_param_compare, fit_param]]
            xdat = dat[fit_param_compare]
            ydat = dat[fit_param]

        xaxis_param_name = _get_param_name(fit_param_compare)
        xaxis_units = _get_param_units(fit_param_compare)
        try:
            xaxis_units = xaxis_units(**kwargs)
        except TypeError:
            pass
        if xaxis_units:
            xaxis_title = '{} ({})'.format(xaxis_param_name, xaxis_units)
        else:
            xaxis_title = xaxis_param_name
        if df_params._drmetric == 'dip':
            xaxis_title = 'DIP {}'.format(xaxis_title)
        else:
            xaxis_title = '{:g} hr viability {}'.format(
                df_params._viability_time,
                xaxis_title)

        range_bounded_params = set()

        for param in (fit_param, fit_param_compare):
            match = IC_REGEX.match(param)
            if match:
                range_bounded_params.add(match.group())
            match = EC_REGEX.match(param)
            if match:
                range_bounded_params.add(match.group())
            match = E_REGEX.match(param)
            if match:
                range_bounded_params.add('ec' + match.groups(0)[0])
            match = E_REL_REGEX.match(param)
            if match:
                range_bounded_params.add('ec' + match.groups(0)[0])

        symbols, hovertext = _symbols_hovertext_two_param_scatter(
            df_params, range_bounded_params)

        xdat_fit = np.log10(xdat) if _param_is_log(fit_param_compare) else xdat
        ydat_fit = np.log10(ydat) if _param_is_log(fit_param) else ydat

        data = []

        if len(xdat_fit) > 0:
            # Remove any infinite or NaN values from best fit line data
            fitdat_mask = (~np.isnan(xdat_fit) & np.isfinite(xdat_fit) &
                           ~np.isnan(ydat_fit) & np.isfinite(ydat_fit) &
                           [s != 'cross' for s in symbols])
            xdat_fit = xdat_fit[fitdat_mask].values
            ydat_fit = ydat_fit[fitdat_mask].values

        if len(xdat_fit) > 0:
            slope, intercept, r_value, p_value, std_err = \
                scipy.stats.linregress(xdat_fit, ydat_fit)

            xfit = (min(xdat_fit), max(xdat_fit))
            yfit = [x * slope + intercept for x in xfit]
            if _param_is_log(fit_param_compare):
                xfit = np.power(10, xfit)
            if _param_is_log(fit_param):
                yfit = np.power(10, yfit)
            data.append(go.Scatter(
                x=xfit,
                y=yfit,
                mode='lines',
                hoverinfo="none",
                line=dict(
                    color="darkorange"
                ),
                name='{} vs {} Linear Fit'.format(xaxis_param_name,
                                                  yaxis_param_name),
                showlegend=False
            ))
            layout['annotations'] = [{
                'x': 0.5, 'y': 1.0, 'xref': 'paper', 'yanchor': 'bottom',
                'yref': 'paper', 'showarrow': False,
                'text': 'R<sup>2</sup>: {:0.4g} '
                        'p-value: {:0.4g} '.format(r_value ** 2, p_value)
            }]

        custom_data = [
            {'c': cl, 'd': dr} for cl, dr in zip(
                df_params.index.get_level_values('cell_line'),
                df_params.index.get_level_values('drug')
            )
        ]

        if color_by:
            for idx, tag_name in enumerate(color_groups):
                location = df_params.index.get_level_values(color_by_col).isin(color_groups[tag_name])
                dat = df_params[location]
                symbols, hovertext = _symbols_hovertext_two_param_scatter(
                    dat, range_bounded_params)

                xdat = dat.loc[:, fit_param_compare]
                ydat = dat.loc[:, fit_param]

                data.append(go.Scatter(
                    x=xdat,
                    y=ydat,
                    hovertext=hovertext,
                    hoverinfo="text+x+y",
                    mode='markers',
                    customdata=custom_data,
                    marker={'symbol': symbols,
                            'color': colours[idx]},
                    name=tag_name
                ))
        else:
            colour_list = [colours[1] if s == 'circle' else 'crimson' for s in
                           symbols]

            data.append(go.Scatter(
                x=xdat,
                y=ydat,
                hovertext=hovertext,
                hoverinfo="text+x+y",
                mode='markers',
                customdata=custom_data,
                marker={'symbol': symbols,
                        'color': colour_list},
                name='{} vs {}'.format(xaxis_param_name,
                                       yaxis_param_name)
            ))

        layout['xaxis'] = {'title': xaxis_title,
                           'type': 'log' if _param_is_log(fit_param_compare)
                           else None}
        layout['hovermode'] = 'closest'
        layout['showlegend'] = color_by is not None
    elif not aggregate_cell_lines and not aggregate_drugs:
        if fit_param_sort is None:
            sort_by = [fit_param, 'label']
        elif fit_param_sort == 'label':
            sort_by = ['label', fit_param]
            fit_param_sort = None
        else:
            sort_by = [fit_param_sort, 'label']
        df_params = df_params.sort_values(
            by=sort_by, na_position='first' if _param_na_first(sort_by[0])
            else 'last')
        groups = df_params['label']
        yvals = df_params[fit_param]

        text = None

        ec_match = None
        for regex in (EC_REGEX, E_REGEX, E_REL_REGEX):
            match = regex.match(fit_param)
            if match:
                ec_match = 'ec' + match.groups(0)[0]
                break

        marker_cols = None
        if not color_by:
            marker_cols = colours[1]

        if ec_match:
            msg = _out_of_range_msg(ec_match)
            if not fit_param.startswith('ec'):
                msg = 'Based on ' + msg
            ec_truncated = is_param_truncated(df_params, ec_match)
            text = [msg if x else '' for x in ec_truncated]
            if not color_by:
                marker_cols = [colours[0] if est else colours[1] for
                               est in ec_truncated]
        elif IC_REGEX.match(fit_param):
            msg = _out_of_range_msg(fit_param)
            ic_truncated = is_param_truncated(df_params, fit_param)
            text = [msg if x else '' for x in ic_truncated]
            if not color_by:
                marker_cols = [colours[0] if est else colours[1] for
                               est in ic_truncated]

        if color_by:
            color_ent = df_params.index.get_level_values(color_by_col)

            marker_cols = []
            for c in color_ent:
                for idx, tag_name in enumerate(color_groups):
                    if c in color_groups[tag_name]:
                        marker_cols.append(colours[idx])
                        break
                else:
                    raise ValueError('Entity not found: {}'.format(c))

        if fit_param_sort is not None:
            na_list = df_params[fit_param_sort].isnull()
            if text is None:
                text = [''] * len(df_params)
            for idx in range(len(df_params)):
                if na_list.iloc[idx]:
                    if text[idx]:
                        text[idx] += '<br>'
                    text[idx] += '{} undefined, sorted by {}'.format(
                        _get_param_name(fit_param_sort),
                        _get_param_name(fit_param)
                    )

        custom_data = [
            {'c': cl, 'd': dr} for cl, dr in zip(
                df_params.index.get_level_values('cell_line'),
                df_params.index.get_level_values('drug')
            )
        ]

        data = [go.Bar(x=groups,
                       y=yvals,
                       text=text,
                       name='',
                       customdata=custom_data,
                       showlegend=False,
                       marker={'color': marker_cols}
                       )]

        layout['annotations'] = []

        if color_by:
            # Nasty cludge to get legend to show, by stacking dummy traces
            # with zero height (Plotly doesn't support legend by colour at
            # this time)
            for idx, tag in enumerate(color_groups.items()):
                tag_name, tag_targets = tag
                tag_targets = set(tag_targets)
                data.append(go.Bar(
                    x=groups,
                    y=[c in tag_targets for c in color_ent],
                    text=tag_name,
                    name=tag_name,
                    hoverinfo='none',
                    showlegend=True,
                    marker={'color': colours[idx]}
                ))

            # Mann Whitney U test w/ cont. correction, two-sided
            if len(color_groups) == 2 and fit_param_sort is None:
                group1 = yvals[(y == colours[0] for y in marker_cols)]
                group2 = yvals[(y != colours[0] for y in marker_cols)]
                # Need more than 20 entries in each group, as per scipy docs
                if len(group1) > 20 and len(group2) > 20:
                    mw_u, mw_p = scipy.stats.mannwhitneyu(
                        group1,
                        group2,
                        use_continuity=True,
                        alternative='two-sided'
                    )
                    if not np.isnan(mw_u):
                        layout['annotations'].append({
                            'x': 0.5, 'y': 0.95, 'xref': 'paper',
                            'yanchor': 'bottom',
                            'yref': 'paper', 'showarrow': False,
                            'text': 'Two-sided Mann-Whitney U: {:.4g} '
                                    'p-value: {:.4g}'.format(
                                mw_u, mw_p)
                        })

        layout['annotations'].extend([
            {'x': x, 'y': 0, 'text': '<em>N/A</em>',
             'textangle': 90,
             'xanchor': 'center', 'yanchor': 'bottom',
             'yref': 'paper',
             'showarrow': False,
             'font': {'color': 'rgba(150, 150, 150, 1)'}}
            for x in groups[yvals.isnull().values]])
        layout.setdefault('xaxis', {})['type'] = 'category'
        layout['barmode'] = 'stack'
        layout['showlegend'] = color_by is not None
    else:
        layout['boxmode'] = 'group'

        if color_by and color_by != 'dataset':
            raise CannotPlotError(
                'Custom color schemes are not currently supported with box '
                'plots. Either disable the coloring, or turn off aggregation.'
            )

        if fit_param_sort == fit_param:
            fit_param_sort = None

        if fit_param_sort is None or fit_param_sort == 'label':
            yvals = df_params.loc[:, [fit_param]].dropna()
        else:
            yvals = df_params.loc[:, [fit_param,
                                      fit_param_sort]]
            yvals.dropna(subset=[fit_param], inplace=True)

        if yvals.empty:
            raise CannotPlotError(
                'The selected cell line/drug combinations have no data '
                'available for the selected parameter(s)')

        if aggregate_cell_lines:
            yvals = _aggregate_by_cell_line(yvals, aggregate_cell_lines,
                                            replace_index=True)

        if aggregate_drugs:
            yvals = _aggregate_by_drug(yvals, aggregate_drugs,
                                       replace_index=True)

        drug_groups = yvals.index.get_level_values('drug').unique()
        cell_line_groups = yvals.index.get_level_values('cell_line').unique()

        if len(cell_line_groups) > 1:
            aggregate_by = ['dataset_id', 'cell_line']
            groups = drug_groups
        else:
            groups = cell_line_groups
            aggregate_by = ['dataset_id', 'drug']

        # Add sort character '' to 'Everything else' tag
        SORT_AT_END_CHAR = ''
        vals = yvals.index.levels[yvals.index.names.index(
            aggregate_by[1])].values
        vals = [SORT_AT_END_CHAR + y if y.startswith('Everything else (')
                else y for y in vals]
        yvals.index = yvals.index.set_levels(vals, level=aggregate_by[1])

        # Sort by median effect per drug set, or cell line set if there's
        # only one drug/drug group
        if fit_param_sort is None or fit_param_sort == 'label':
            if fit_param_sort == 'label':
                sort_cols = aggregate_by + ['median']
            else:
                sort_cols = ['median'] + aggregate_by
            yvals['median'] = yvals[fit_param].groupby(
                level=aggregate_by).transform(np.nanmedian)
            yvals.set_index('median', append=True, inplace=True)
            yvals.sort_index(level=sort_cols, ascending=True,
                             inplace=True)
            yvals.reset_index('median', drop=True, inplace=True)
        else:
            # Sort by fit_column_sort, with tie breakers determined by
            # fit_param
            median_cols = yvals.loc[:, [fit_param_sort, 'label']].groupby(
                level=aggregate_by).transform(np.nanmedian)
            median_cols.rename(columns={fit_param_sort: 'median',
                                        'label': 'median2'},
                               inplace=True)
            yvals = pd.concat([yvals, median_cols], axis=1)
            yvals.set_index(['median', 'median2'], append=True, inplace=True)
            yvals.sort_index(level=['median', 'median2'] + aggregate_by,
                             ascending=True, inplace=True)
            yvals.reset_index(['median', 'median2'], drop=True, inplace=True)

        # Remove sort character '' from 'Everything else' tag
        vals = yvals.index.levels[yvals.index.names.index(
            aggregate_by[1])].values
        vals = [y.replace(SORT_AT_END_CHAR, '') for y in vals]
        yvals.index = yvals.index.set_levels(vals, level=aggregate_by[1])

        # Convert yvals to a series
        yvals = yvals.iloc[:, 0]

        data = []
        aggregate_by.remove('dataset_id')
        aggregate_by = aggregate_by[0]

        datasets = yvals.index.get_level_values('dataset_id').unique()

        group_by = ['dataset_id'] if len(datasets) > 1 else []
        group_by += ['drug'] if aggregate_by == 'cell_line' else ['cell_line']

        for grp_name, grp in yvals.groupby(level=group_by):
            if len(groups) > 1:
                group_name = grp_name if isinstance(grp_name, str) else \
                    "<br>".join(str(g) for g in grp_name)
            else:
                # If there's only one drug/cell line group, just need dataset
                group_name = grp_name[0] if grp_name else None
            data.append(go.Box(x=grp.index.get_level_values(aggregate_by),
                               y=grp,
                               name=group_name
                               ))

        layout['annotations'] = []
        if len(groups) == 1:
            annotation_label = str(groups[0])
            layout['annotations'].append({
                'x': 0.5, 'y': 1.0, 'xref': 'paper', 'yanchor': 'bottom',
                'yref': 'paper', 'showarrow': False, 'text': annotation_label
            })

        # One way anova test
        anova_f, anova_p = scipy.stats.f_oneway(
            *[x[1].values for x in yvals.groupby(level=aggregate_by)]
        )
        if not np.isnan(anova_f):
            layout['annotations'].append({
                'x': 0.5, 'y': 0.95, 'xref': 'paper',
                'yanchor': 'bottom',
                'yref': 'paper', 'showarrow': False,
                'text': 'One-way ANOVA F: {:.4g} p-value: {:.4g}'.format(
                    anova_f, anova_p)
            })

    layout = go.Layout(layout)

    return go.Figure(data=data, layout=layout)


def _aggregate_by_drug(yvals, aggregate_drugs, replace_index=True):
    return _aggregate_by_tag(yvals, aggregate_drugs, 'drug',
                             replace_index=replace_index)


def _aggregate_by_cell_line(yvals, aggregate_cell_lines,
                            replace_index=True):
    return _aggregate_by_tag(yvals, aggregate_cell_lines,
                             'cell_line', replace_index=replace_index)


def _aggregate_by_tag(yvals, aggregate_items, label_type,
                      replace_index=True, add_counts=True):
    if aggregate_items in (None, False):
        return yvals

    if aggregate_items is True:
        items = yvals.index.get_level_values(
            level=label_type).unique().tolist()
        aggregate_items = {
            _create_label_max_items(items, 5): items
        }

    new = pd.DataFrame()

    label_type_tag = label_type + '_tag'

    for tag_name, names in aggregate_items.items():
        yvals_tmp = yvals.loc[yvals.index.isin(names, level=label_type), :]
        # Avoid warning about setting on copy, since we're using the copy to
        # build a new dataframe
        yvals_tmp.is_copy = None

        # Add counts to the tag names
        if add_counts:
            tag_name = '{} ({})'.format(tag_name, len(
                yvals_tmp.index.get_level_values(label_type).unique()))

        yvals_tmp[label_type_tag] = np.repeat(tag_name, len(yvals_tmp))
        new = new.append(yvals_tmp)

    labels = list(new.index.names)
    new.reset_index([l for l in labels if l != label_type], inplace=True)
    labels[labels.index(label_type)] = label_type_tag
    new.set_index(labels, inplace=True, drop=replace_index)
    if replace_index:
        new.index.rename(label_type, level=label_type_tag, inplace=True)

    return new


def _create_label_max_items(items, max_items=5):
    n = len(items)
    if n > max_items:
        items = items[0:max_items]
    annotation_label = ", ".join(items)
    if n > max_items:
        annotation_label += " and {} more".format(n - len(items))
    return annotation_label


def plot_time_course(hts_pandas,
                     log_yaxis=False, assay_name='Assay', title=None,
                     subtitle=None, show_dip_fit=False,
                     template=config.plotly_template):
    """
    Plot a dose response time course

    Parameters
    ----------
    hts_pandas: HtsPandas
        Dataset containing a single cell line/drug combination
    log_yaxis: bool
        Use log scale on y-axis
    assay_name: str
        The name of the assay to use for the time course (only used for
        multi-assay datasets)
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    show_dip_fit: bool
        Overlay the DIP rate fit on the time course
    template: str
        Name of plotly template (https://plot.ly/python/templates/)

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    if show_dip_fit and not log_yaxis:
        raise ValueError('log_yaxis must be True when show_dip_fit is True')

    df_doses = hts_pandas.doses
    if hts_pandas.controls is not None:
        df_controls = hts_pandas.controls
    else:
        df_controls = None
    df_vals = hts_pandas.assays

    df_assays_avail = hts_pandas.assay_names
    if len(df_assays_avail) == 1:
        assay = df_assays_avail[0]
    elif assay_name in df_assays_avail:
        assay = assay_name
    else:
        raise ValueError('{} is not a valid assay. Options are {}'.format(
            assay_name, df_assays_avail))

    if df_controls is not None:
        if 'dataset' in df_controls.index.names:
            dsets = df_controls.index.get_level_values('dataset').unique()
            if len(dsets) > 1:
                raise ValueError('Multiple control datasets present. '
                                 'Plotting a time course requires a single '
                                 'dataset.')
            df_controls = df_controls.loc[dsets[0]]
        df_controls = df_controls.loc[assay]

        # Only use controls from the plates in the expt data
        plates = df_doses['plate'].unique()
        df_controls = df_controls.loc[(slice(None), plates), :]
    df_vals = df_vals.loc[assay]

    if len(hts_pandas.drugs) > 1 or len(hts_pandas.cell_lines) > 1:
        raise ValueError('Dataset has multiple drugs and/or cell lines. Time '
                         'courses are currently only available for single '
                         'drug/cell line combinations.')

    traces = []
    traces_fits = []

    if title is None:
        title = _make_title('Time course', df_doses)
    title = _combine_title_subtitle(title, subtitle)

    colours = _sns_to_rgb(sns.color_palette(
        "husl", len(df_doses.index.get_level_values(level='dose').unique())))

    ctrl_color = 'white' if template == 'plotly_dark' else 'black'

    if show_dip_fit:
        if df_controls is not None:
            dip_rate_ctrl = ctrl_dip_rates(df_controls)
            dip_rate_ctrl.index = dip_rate_ctrl.index.droplevel(level='cell_line')
        dip_rates = expt_dip_rates(df_doses, df_vals)
        dip_rates.reset_index(inplace=True)
        dip_rates.set_index('well_id', inplace=True)

    # Controls
    if df_controls is not None:
        is_first_control = True
        for well_id, timecourse in df_controls.groupby(level='well_id'):
            timecourse = timecourse['value']
            t0_offset = 0
            if log_yaxis:
                timecourse = np.log2(timecourse)
                t0_offset = timecourse[0]
                timecourse -= t0_offset
            x_range = [t.total_seconds() / SECONDS_IN_HOUR for t in
                       timecourse.index.get_level_values('timepoint')]
            traces.append(go.Scatter(
                x=x_range,
                y=timecourse,
                mode='lines+markers',
                line={'color': ctrl_color,
                      'shape': 'spline',
                      'dash': 'dot' if show_dip_fit else None},
                marker={'size': 5},
                name='Control',
                legendgroup='__Control',
                showlegend=is_first_control
            ))
            is_first_control = False

            if show_dip_fit and df_controls is not None:
                dip_well = dip_rate_ctrl.loc[
                    dip_rate_ctrl.index.get_level_values('well_id') == well_id]
                minmax = [np.min(x_range), np.max(x_range)]

                dip_points = [x * dip_well['dip_rate'] +
                              dip_well['dip_y_intercept'] - t0_offset
                              for x in minmax]

                traces_fits.append(go.Scatter(
                    x=minmax,
                    y=dip_points,
                    mode='lines',
                    line={'color': ctrl_color},
                    marker={'size': 5},
                    name='DIP fit Control',
                    legendgroup='__Control',
                    showlegend=False
                ))

    # Experiment (non-control)
    for dose, wells in df_doses.groupby(level='dose'):
        this_colour = colours.pop()

        for well_idx, well_id in enumerate(wells['well_id']):
            try:
                timecourse = df_vals.loc[well_id]['value']
            except KeyError:
                continue
            t0_offset = 0
            if log_yaxis:
                timecourse = np.log2(timecourse)
                t0_offset = timecourse[0]
                timecourse -= t0_offset
            x_range = [t.total_seconds() / SECONDS_IN_HOUR for t in
                       timecourse.index.get_level_values('timepoint')]
            dose_str = format_dose(dose, array_as_string=" &amp; ")
            traces.append(go.Scatter(
                x=x_range,
                y=timecourse,
                mode='lines+markers',
                line={'color': this_colour,
                      'shape': 'spline',
                      'dash': 'dot' if show_dip_fit else None},
                marker={'size': 5},
                name=dose_str,
                customdata=({'csvname': dose}, ),
                legendgroup=dose_str,
                showlegend=well_idx == 0
            ))

            if show_dip_fit:
                dip_well = dip_rates.loc[well_id]
                minmax = [dip_well['dip_first_timepoint'], np.max(x_range)]
                dip_points = [x*dip_well['dip_rate'] +
                              dip_well['dip_y_intercept'] - t0_offset
                              for x in minmax]

                traces_fits.append(go.Scatter(
                    x=minmax,
                    y=dip_points,
                    mode='lines',
                    line={'color': this_colour},
                    marker={'size': 5},
                    name=dose_str,
                    customdata=({'csvname': 'DIP fit ' + dose_str}, ),
                    legendgroup=dose_str,
                    showlegend=False
                ))

    data = (traces + traces_fits)
    if log_yaxis:
        assay_name = "Change in log<sub>2</sub> {}".format(assay_name)
    max_time = df_vals.index.get_level_values('timepoint').max()
    layout = go.Layout(title=title,
                       xaxis={'title': 'Time (hours)',
                              'range': (0, 120) if max_time <=
                                       np.timedelta64(120, 'h') else None,
                              'dtick': 12},
                       yaxis={'title': assay_name,
                              'range': (-2, 7) if log_yaxis else None},
                       template=template
                       )
    return go.Figure(data=data, layout=layout)


def plot_ctrl_dip_by_plate(df_controls, title=None, subtitle=None,
                           template=config.plotly_template):
    """

    Parameters
    ----------
    df_controls: pd.DataFrame
        Control well DIP values
    title: str, optional
        Title (or None to auto-generate)
    subtitle: str, optional
        Subtitle (or None to auto-generate)
    template: str
        Name of plotly template (https://plot.ly/python/templates/)

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    # Sort by median DIP rate
    df_controls = df_controls.copy()
    df_controls['cl_median'] = df_controls['dip_rate'].groupby(
        level=['cell_line']).transform(np.nanmedian)
    df_controls['plate_median'] = df_controls['dip_rate'].groupby(
        level=['cell_line', 'plate']).transform(np.nanmedian)
    df_controls.sort_values(by=['cl_median', 'plate_median'], inplace=True)

    if title is None:
        title = 'Control DIP rates by plate'

    if 'dataset' in df_controls.index.names:
        dataset_names = df_controls.index.get_level_values('dataset').unique()

        if len(dataset_names) != 1:
            raise ValueError('This function can only plot controls from a '
                             'single dataset')

        if subtitle is None:
            subtitle = dataset_names[0]

    title = _combine_title_subtitle(title, subtitle)

    traces = []
    for grp, ctrl_dat in df_controls.groupby(level=['cell_line']):
        traces.append(go.Box(
            x=ctrl_dat.index.get_level_values('plate').values,
            y=ctrl_dat['dip_rate'].values,
            name=grp
        ))

    layout = go.Layout(title=title,
                       template=template,
                       yaxis={'title': 'DIP Rate (h<sup>-1</sup>)'})
    return go.Figure(data=traces, layout=layout)


def plot_plate_map(plate_data, color_by='dip_rates',
                   missing_color='lightgray', subtitle=None,
                   template=config.plotly_template):
    """

    Parameters
    ----------
    plate_data: thunor.io.PlateData
        Plate map layout data
    color_by: str
        Attribute to color wells by, must be numerical (default: dip_rates)
    missing_color: str
        Color to use for missing values (default: lightgray)
    subtitle: str or None
        Subtitle, or None to auto-generate
    template: str
        Name of plotly template (https://plot.ly/python/templates/)

    Returns
    -------
    plotly.graph_objs.Figure
        A plotly figure object containing the graph
    """
    maintitle = 'DIP Rate Plate Map'
    if subtitle is None:
        subtitle = 'Plate {}'.format(plate_data.plate_name)
        if plate_data.dataset_name:
            subtitle += ' ({})'.format(plate_data.dataset_name)
    title = _combine_title_subtitle(maintitle, subtitle)

    well_color_basis = np.array(getattr(plate_data, color_by),
                                dtype=np.double)

    well_max = np.nanmax(well_color_basis)
    well_min = np.nanmin(well_color_basis)
    pos_pal = sns.light_palette("#f48000", as_cmap=True)
    neg_pal = sns.light_palette("#3f83a3", as_cmap=True)

    well_color = _sns_to_rgb([
        missing_color
        if np.isnan(val)
        else pos_pal(val / well_max)
        if val > 0
        else neg_pal(val / well_min)
        for val in well_color_basis
    ])

    cols = plate_data.width
    rows = plate_data.height
    num_wells = cols * rows

    well_rad = PLATE_MAP_WELL_DIAM / 2

    col_labels = [str(i + 1) for i in range(cols)]
    row_labels = [chr(ASCII_CAP_A + i) for i in range(rows)]

    well_shapes = [{
        'opacity': 1.0,
        'xref': 'x',
        'yref': 'y',
        'fillcolor': well_color[well_num],
        'x0': (well_num % cols),
        'y0': rows - (well_num // cols),
        'x1': (well_num % cols) + PLATE_MAP_WELL_DIAM,
        'y1': rows - (well_num // cols) + PLATE_MAP_WELL_DIAM,
        'type': 'circle',
        'line': {
            'color': 'darkgray',
            'width': 0.5
        }
    } for well_num in range(num_wells)]

    well_objs = go.Scatter(
        x=[(well_num % cols) + well_rad for well_num in range(num_wells)],
        y=[rows - (well_num // cols) + well_rad for well_num in
           range(num_wells)],
        text='',
        hovertext=['Well {}{}<br>'
                   'DIP: {}<br>'
                   'Cell Line: {}<br>'
                   'Drug: {}<br>'
                   'Dose: {}'.format(
                        row_labels[well_num // cols],
                        col_labels[well_num % cols],
                        plate_data.dip_rates[well_num],
                        plate_data.cell_lines[well_num],
                        " &amp; ".join(["(None)" if d is None else d for d in
                                        plate_data.drugs[well_num]]) if
                                        plate_data.drugs[well_num] else 'None',
                        " &amp; ".join([format_dose(d) for d in
                                        plate_data.doses[well_num]]) if
                                        plate_data.doses[well_num] else 'N/A'
            )
            for well_num in range(num_wells)],
        hoverinfo='text',
        mode='text'
    )

    label_color = 'white' if template == 'plotly_dark' else 'black'

    col_labels = go.Scatter(
        x=[i + well_rad for i in range(cols)],
        y=[rows + 1.2] * cols,
        mode='text',
        text=col_labels,
        hoverinfo='none',
        textfont=dict(
            color=label_color,
            # size=18,
        )
    )

    row_labels = go.Scatter(
        x=[-well_rad] * rows,
        y=[i + 1 + well_rad for i in reversed(range(rows))],
        mode='text',
        text=row_labels,
        hoverinfo='none',
        textfont=dict(
            color=label_color,
            # size=18,
        )
    )

    data = (well_objs, col_labels, row_labels)

    layout = go.Layout({
        'xaxis': {
            'showticklabels': False,
            'showgrid': False,
            'zeroline': False,
        },
        'yaxis': {
            'showticklabels': False,
            'showgrid': False,
            'zeroline': False,
            'scaleanchor': 'x'
        },
        'shapes': well_shapes,
        'margin': {
            'l': 0,
            'r': 0,
            'b': 0,
            't': 50
        },
        'hovermode': 'closest',
        'showlegend': False,
        'title': title,
        'template': template
    })

    return go.Figure(data=data, layout=layout)



from datetime import timedelta

SECONDS_IN_HOUR = 3600


def viability(df_data, time_hrs=72, assay_name=None, include_controls=True):
    """
    Calculate viability at the specified time point

    Viability is calculated as the assay value over the mean of controls
    from the same plate, cell line, and time point

    Parameters
    ----------
    df_data: HtsPandas
        HTS dataset
    time_hrs: float
        Time in hours to use for viability. The closest time point in each
        well to the one specified is used.
    assay_name: str, optional
        The assay name to use for viability calculation, or None to use the
        default proliferation assay
    include_controls: bool
        Return the control values for reference as a the second entry in a
        two-tuple, if True

    Returns
    -------
    pd.DataFrame, pd.Series or None
        A DataFrame containing the viability results and a Series containing
        the control values, if requested (None is returned as the second
        return value otherwise)
    """
    if df_data.controls is None:
        raise ValueError('Control wells not found, and are needed for '
                         'viability calculation')

    time = timedelta(hours=time_hrs)

    # Select assay per dataset
    if 'dataset' in df_data.doses.index.names:
        assays = df_data.assays.reset_index('assay')
    else:
        if assay_name is None:
            assay_name = df_data.dip_assay_name
        assays = df_data.assays.loc[assay_name]

    # Filter assays by nearest timepoint
    assays = _get_closest_timepoint_for_each_well(assays, time)
    assays.reset_index('timepoint', inplace=True)

    # Merge counts with well annotations
    df = df_data.doses.merge(assays, left_on='well_id', right_index=True)

    if 'dataset' in df_data.doses.index.names:
        dataset_assays = df.groupby('dataset')['assay'].unique()
        if not (dataset_assays.apply(len) == 1).all():
            raise NotImplementedError('Cannot calculate viability across two '
                                      'datasets when the datasets are '
                                      'multi-assay')
    if 'dataset' in df_data.controls.index.names:
        controls = df_data.controls.reset_index('assay', drop=True)
    else:
        controls = df_data.controls.loc[assay_name]
    controls = _get_closest_timepoint_for_each_well(controls, time)

    # Get average control cell count on each plate for each time point
    idx_cols = ['plate', 'cell_line', 'timepoint']
    if 'dataset' in df.index.names:
        idx_cols = ['dataset'] + idx_cols

    controls_means = controls['value'].groupby(level=idx_cols).mean()

    if include_controls:
        controls['value'] = controls['value'].groupby(level=idx_cols).apply(
            lambda x: x / x.mean())

    df.reset_index(inplace=True)
    df.rename(columns={'plate_id': 'plate'}, inplace=True)
    df.set_index(idx_cols, inplace=True)

    df = df.join(controls_means.to_frame(), rsuffix='_ctrl')
    df.reset_index(inplace=True)
    final_idx_cols = ['drug', 'cell_line', 'dose', 'well_id']
    if 'dataset' in df.columns:
        final_idx_cols = ['dataset'] + final_idx_cols
    df.set_index(final_idx_cols, inplace=True)

    df['viability'] = df['value'] / df['value_ctrl']

    timepoints = df['timepoint'].unique()
    if len(timepoints) == 1:
        time_hrs = timepoints[0].astype('timedelta64[h]').item()\
                       .total_seconds() / SECONDS_IN_HOUR

    df._viability_time = time_hrs
    df._viability_assay = assay_name

    if include_controls:
        return df, controls['value']
    else:
        return df, None


def _get_closest_timepoint_for_each_well(dataframe, timediff):
    dataframe['timediff'] = abs(dataframe.index.get_level_values('timepoint') -
                                timediff)
    dataframe = dataframe.loc[dataframe.groupby('well_id')[
        'timediff'].idxmin()]

    return dataframe



import numpy as np
import scipy.optimize
import scipy.stats
from abc import abstractmethod
import pandas as pd
from decimal import Decimal, Overflow
import warnings

PARAM_EQUAL_ATOL = 1e-16
PARAM_EQUAL_RTOL = 1e-12
# Curve fitting E0 tolerance parameters
E0_TEST_MIN_CTRLS_STD_DEV = 5  # min # controls to use std. dev.
E0_TEST_MULTIPLIER = 1.2
# Value to divide minimum dose by to get "control dose" for visualization
# and dip fits
CTRL_DOSE_DIVISOR = 10.0


class ValueWarning(UserWarning):
    pass


class AUCFitWarning(ValueWarning):
    pass


class AAFitWarning(ValueWarning):
    pass


class DrugCombosNotImplementedError(NotImplementedError):
    """ This function does not support drug combinations yet """
    pass


class HillCurve(object):
    """ Base class defining Hill/log-logistic curve functionality """
    fit_bounds = (np.NINF, np.PINF)
    null_response_fn = np.mean
    max_fit_evals = None

    def __init__(self, popt):
        self.popt = popt

    @classmethod
    @abstractmethod
    def fit_fn(cls, x, *params):
        pass

    def fit(self, x):
        return self.fit_fn(x, *self.popt)

    def fit_rel(self, x):
        return None

    @classmethod
    @abstractmethod
    def initial_guess(cls, x, y):
        pass

    @abstractmethod
    def divisor(self):
        pass

    @abstractmethod
    def ec50(self):
        pass

    @abstractmethod
    def e0(self):
        pass

    @abstractmethod
    def emax(self):
        pass

    @abstractmethod
    def hill_slope(self):
        pass


class HillCurveNull(HillCurve):
    @classmethod
    def fit_fn(cls, x, ymean):
        return ymean

    def fit(self, x):
        return self.fit_fn(x, self.popt)

    @classmethod
    def initial_guess(cls, x, y):
        return np.mean(y)

    @property
    def divisor(self):
        return self.popt

    def ic(self, ic_num=50):
        return None

    def ec(self, ec_num=50):
        return None

    @property
    def ec50(self):
        return None

    @property
    def e0(self):
        return self.popt

    @property
    def emax(self):
        return self.popt

    @property
    def hill_slope(self):
        return None

    def auc(self, *args, **kwargs):
        return 0.0

    def aa(self, *args, **kwargs):
        return 0.0


class HillCurveLL4(HillCurve):
    max_fit_evals = 10000

    def __init__(self, popt):
        super(HillCurveLL4, self).__init__(popt)
        self._popt_rel = None

    @classmethod
    def fit_fn(cls, x, b, c, d, e):
        """
        Four parameter log-logistic function ("Hill curve")

        Parameters
        ----------
        x: np.ndarray
            One-dimensional array of "x" values
        b: float
            Hill slope
        c: float
            Maximum response (lower plateau)
        d: float
            Minimum response (upper plateau)
        e: float
            EC50 value

        Returns
        -------
        np.ndarray
            Array of "y" values using the supplied curve fit parameters on "x"
        """
        return c + (d - c) / (1 + np.exp(b * (np.log(x) - np.log(e))))

    @classmethod
    def initial_guess(cls, x, y):
        """
        Heuristic function for initial fit values

        Uses the approach followed by R's drc library:
        https://cran.r-project.org/web/packages/drc/index.html

        Parameters
        ----------
        x: np.ndarray
            Array of "x" (dose) values
        y: np.ndarray
            Array of "y" (response) values

        Returns
        -------
        list
            Four-valued list corresponding to initial estimates of the
            parameters defined in the :func:`ll4` function.
        """
        c_val, d_val = _find_cd_ll4(y)
        b_val, e_val = _find_be_ll4(x, y, c_val, d_val)

        return b_val, c_val, d_val, e_val

    @property
    def ec50(self):
        return self.popt[3]

    @property
    def e0(self):
        return self.popt[2]

    @property
    def emax(self):
        return self.popt[1]

    @property
    def hill_slope(self):
        return self.popt[0]

    def ic(self, ic_num=50):
        """
        Find the inhibitory concentration value (e.g. IC50)

        Parameters
        ----------
        ic_num: int
            IC number between 0 and 100 (response level)

        Returns
        -------
        float
            Inhibitory concentration value for requested response value
        """
        emax = self.emax
        if not isinstance(emax, float):
            return None
        e0 = self.e0
        if emax > e0:
            emax, e0 = e0, emax

        ic_frac = ic_num / 100.0

        icN = self.ec50 * (ic_frac / (1 - ic_frac - (emax / e0))) ** (
                           1 / self.hill_slope)

        # Overflow will lead to -inf, which we deal with here
        if np.isnan(icN) or np.isinf(icN):
            icN = None

        return icN

    def ec(self, ec_num=50):
        """
        Find the effective concentration value (e.g. IC50)

        Parameters
        ----------
        ec_num: int
            EC number between 0 and 100 (response level)

        Returns
        -------
        float
            Effective concentration value for requested response value
        """
        if ec_num >= 100:
            return None

        ec_frac = ec_num / 100.0

        return self.ec50 * (ec_frac / (1 - ec_frac)) ** (1 / self.hill_slope)

    def auc(self, min_conc):
        """
        Find the area under the curve

        Parameters
        ----------
        min_conc: float
            Minimum concentration to consider for fitting the curve

        Returns
        -------
        float
            Area under the curve (AUC) value
        """
        emax = self.emax
        if not isinstance(emax, float):
            return None
        e0 = self.e0
        if emax > e0:
            # TODO: Calculate AUC for ascending curves
            return None

        min_conc_hill = min_conc ** self.hill_slope
        return (np.log10(
            (self.ec50 ** self.hill_slope + min_conc_hill) / min_conc_hill) /
                self.hill_slope) * ((e0 - emax) / e0)

    def aa(self, min_conc, max_conc):
        """
        Find the activity area (area over the curve)

        Parameters
        ----------
        min_conc: float
            Minimum concentration to consider for fitting the curve
        max_conc: float
            Maximum concentration to consider for fitting the curve

        Returns
        -------
        float
            Activity area value
        """
        emax = self.emax
        if not isinstance(emax, float):
            return None
        e0 = self.e0
        if emax > e0:
            # TODO: Calculate AA for ascending curves
            return None

        hill = Decimal(self.hill_slope)
        try:
            ec50_hill = Decimal(self.ec50) ** hill
            min_conc = Decimal(min_conc)
            max_conc = Decimal(max_conc)

            return np.float64(((ec50_hill + max_conc ** hill).log10()
                               - (ec50_hill + min_conc ** hill).log10())
                              / hill) * ((e0 - emax) / e0)
        except Overflow:
            warnings.warn('Overflow finding activity area '
                          '(Hill coeff.: {:.4g})'.format(hill))
            return None

    @property
    def divisor(self):
        return max(self.emax, self.e0)

    @property
    def popt_rel(self):
        if self._popt_rel is None:
            self._popt_rel = self.popt.copy()
            self.popt_rel[2] /= self.divisor
            self.popt_rel[1] /= self.divisor
        return self._popt_rel

    def fit_rel(self, x):
        return self.fit_fn(x, *self.popt_rel)


class HillCurveLL3u(HillCurveLL4):
    """ Three parameter log logistic curve, for viability data """
    # Constrain 0<=emax<=1, Hill slope +ve
    fit_bounds = (
        (0.0, 0.0, np.NINF),
        (np.PINF, 1.0, np.PINF)
    )
    max_fit_evals = None

    @staticmethod
    def null_response_fn(_):
        return np.float64(1.0)

    @classmethod
    def fit_fn(cls, x, b, c, e):
        """
        Three parameter log-logistic function ("Hill curve")

        Parameters
        ----------
        x: np.ndarray
            One-dimensional array of "x" values
        b: float
            Hill slope
        c: float
            Maximum response (lower plateau)
        e: float
            EC50 value

        Returns
        -------
        np.ndarray
            Array of "y" values using the supplied curve fit parameters on "x"
        """
        return super(HillCurveLL3u, cls).fit_fn(x, b, c, 1.0, e)

    @classmethod
    def initial_guess(cls, x, y):
        hill, emax, _, ec50 = super().initial_guess(x, y)
        if emax < 0.0:
            emax = 0.0
        elif emax > 1.0:
            emax = 1.0
        return hill, emax, ec50

    @property
    def ec50(self):
        return self.popt[2]

    @property
    def e0(self):
        return 1.0

    @property
    def popt_rel(self):
        if self._popt_rel is None:
            self._popt_rel = self.popt.copy()
            self._popt_rel[1] /= self.divisor
        return self._popt_rel


class HillCurveLL2(HillCurveLL3u):
    @classmethod
    def fit_fn(cls, x, b, e):
        """
        Two parameter log-logistic function ("Hill curve")

        Parameters
        ----------
        x: np.ndarray
            One-dimensional array of "x" values
        b: float
            Hill slope
        e: float
            EC50 value

        Returns
        -------
        np.ndarray
            Array of "y" values using the supplied curve fit parameters on "x"
        """
        return super(HillCurveLL3u, cls).fit_fn(x, b, 0.0, 1.0, e)

    @classmethod
    def initial_guess(cls, x, y):
        b, _, _, e = super(HillCurveLL3u, cls).initial_guess(x, y)
        return b, e

    @property
    def ec50(self):
        return self.popt[1]

    @property
    def e0(self):
        return 1.0

    @property
    def emax(self):
        return 0.0

    @property
    def popt_rel(self):
        if self._popt_rel is None:
            self._popt_rel = self.popt.copy()
        return self._popt_rel


def fit_drc(doses, responses, response_std_errs=None, fit_cls=HillCurveLL4,
            null_rejection_threshold=0.05,
            ctrl_dose_test=False):
    """
    Fit a dose response curve

    Parameters
    ----------
    doses: np.ndarray
        Array of dose values
    responses: np.ndarray
        Array of response values, e.g. viability, DIP rates
    response_std_errs: np.ndarray, optional
        Array of fit standard errors for the response values
    fit_cls: Class
        Class to use for fitting (default: 4 parameter log logistic
        "Hill" curve)
    null_rejection_threshold: float, optional
        p-value for rejecting curve fit against no effect "flat" response
        model by F-test (default: 0.05). Set to None to skip test.
    ctrl_dose_test: boolean
        If True, the minimum dose is assumed to represent control values (in
        DIP rate curves), and will reject fits where E0 is greater than a
        standard deviation higher than the mean of the control response values.
        Leave as False to skip the test.

    Returns
    -------
    HillCurve
        A HillCurve object containing the fit parameters
    """
    # Remove any NaNs
    response_nans = np.isnan(responses)
    if np.any(response_nans):
        doses = doses[~response_nans]
        responses = responses[~response_nans]
        if response_std_errs is not None:
            response_std_errs = response_std_errs[~response_nans]

    # Sort the data - curve fit can be sensitive to order!
    try:
        if response_std_errs is None:
            doses, responses = zip(*sorted(zip(doses, responses)))
        else:
            doses, responses, response_std_errs = zip(*sorted(zip(
                doses, responses, response_std_errs)))
    except ValueError:
        # Occurs when doses/responses is empty
        return None

    curve_initial_guess = fit_cls.initial_guess(doses, responses)
    try:
        popt, pcov = scipy.optimize.curve_fit(fit_cls.fit_fn,
                                              doses,
                                              responses,
                                              bounds=fit_cls.fit_bounds,
                                              p0=curve_initial_guess,
                                              sigma=response_std_errs,
                                              maxfev=fit_cls.max_fit_evals
                                              )
    except RuntimeError:
        # Some numerical issue with curve fitting
        return None
    except ValueError:
        return None
    except TypeError as te:
        # This occurs if there are fewer data points than parameters
        te_str = str(te)
        if 'Improper input:' in te_str:
            warnings.warn(te_str)
            return None
        else:
            raise

    if any(np.isnan(popt)):
        # Ditto
        return None

    fit_obj = fit_cls(popt)

    if null_rejection_threshold is not None:
        null_response_value = fit_cls.null_response_fn(responses)
        response_curve = fit_obj.fit(doses)

        # F test vs flat linear "no effect" fit
        ssq_model = ((response_curve - responses) ** 2).sum()
        ssq_null = ((null_response_value - responses) ** 2).sum()

        df = len(doses) - len(popt)

        f_ratio = (ssq_null-ssq_model)/(ssq_model/df)
        p = 1 - scipy.stats.f.cdf(f_ratio, 1, df)

        if p > null_rejection_threshold:
            return HillCurveNull(null_response_value)

    if fit_obj.ec50 < np.min(doses):
        # Reject fit if EC50 less than min dose
        return None

    if ctrl_dose_test:
        responses = np.array(responses)
        controls = responses[np.equal(doses, np.min(doses))]
        if len(controls) >= E0_TEST_MIN_CTRLS_STD_DEV:
            ctrl_resp_thresh = np.mean(controls) + np.std(controls)
        else:
            ctrl_resp_thresh = E0_TEST_MULTIPLIER * np.mean(controls)
        if fit_obj.e0 > ctrl_resp_thresh:
            return None

    return fit_obj


def _response_transform(y, c_val, d_val):
    return np.log((d_val - y) / (y - c_val))


def _find_be_ll4(x, y, c_val, d_val, slope_scaling_factor=1,
                 dose_transform=np.log,
                 dose_inv_transform=np.exp):
    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(
        dose_transform(x),
        _response_transform(y, c_val, d_val)
    )
    b_val = slope_scaling_factor * slope
    e_val = dose_inv_transform(-intercept / (slope_scaling_factor * b_val))

    return b_val, e_val


def _find_cd_ll4(y, scale=0.001):
    ymin = np.min(y)
    ymax = np.max(y)
    len_y_range = scale * (ymax - ymin)

    return ymin - len_y_range, ymax + len_y_range


def _get_control_responses(ctrl_dip_data, dataset, cl_name, dip_grp):
    if ctrl_dip_data is None:
        return None

    if dataset is not None and 'dataset' in ctrl_dip_data.index.names:
        ctrl_dip_data_cl = ctrl_dip_data.loc[dataset]
    else:
        ctrl_dip_data_cl = ctrl_dip_data

    ctrl_dip_data_cl = ctrl_dip_data_cl.loc[cl_name]

    # Only use controls from the same plates as the expt
    if 'plate' in dip_grp.index.names:
        plates = dip_grp.index.get_level_values('plate').unique().values
    elif 'plate' in dip_grp.columns:
        plates = dip_grp['plate'].unique()
    else:
        plates = []

    ctrl_dip_data_cl = ctrl_dip_data_cl.loc[ctrl_dip_data_cl.index.isin(
        plates, level='plate')]

    if ctrl_dip_data_cl.empty:
        return None

    return ctrl_dip_data_cl


# Parameter fitting section #
def aa_obs(responses, doses=None):
    """
    Activity Area (observed)

    Parameters
    ----------
    responses: np.array or pd.Series
        Response values, with dose values in the Index if a Series is
        supplied
    doses: np.array or None
        Dose values - only required if responses is not a pd.Series

    Returns
    -------
    float
        Activity area (observed)
    """
    if doses is None:
        responses = responses.groupby('dose').agg('mean')
        doses = responses.index.get_level_values('dose')
    responses_shifted = 1.0 - np.minimum(responses, 1.0)
    # Ensure array is sorted by dose
    positions = np.argsort(doses)
    responses_shifted = responses_shifted[positions]
    doses = doses[positions]
    return np.trapz(responses_shifted, np.log10(doses))


def fit_params_minimal(ctrl_data, expt_data,
                       fit_cls=HillCurveLL4,
                       ctrl_dose_fn=lambda doses: np.min(doses) /
                       CTRL_DOSE_DIVISOR):
    """
    Fit dose response curves to DIP or viability, and calculate statistics

    This function only fits curves and stores basic fit parameters. Use
    :func:`fit_params` for more statistics and parameters.

    Parameters
    ----------
    ctrl_data: pd.DataFrame or None
        Control DIP rates from :func:`dip_rates` or :func:`ctrl_dip_rates`.
        Set to None to not use control data.
    expt_data: pd.DataFrame
        Experiment (non-control) DIP rates from :func:`dip_rates` or
        :func:`expt_dip_rates`
    fit_cls: Class
        Class to use for curve fitting (default: :func:`HillCurveLL4`)
    ctrl_dose_fn: function
        Function to use to set an effective "dose" (non-zero) for controls.
        Takes the list of experiment doses as an argument.

    Returns
    -------
    pd.DataFrame
        DataFrame containing DIP rate curve fits and parameters
    """
    expt_data_orig = expt_data

    if 'dataset' in expt_data.index.names:
        datasets = expt_data.index.get_level_values('dataset').unique()
    else:
        datasets = None
    cell_lines = expt_data.index.get_level_values('cell_line').unique()
    drugs = expt_data.index.get_level_values('drug').unique()

    has_drug_combos = drugs.map(len).max() > 1
    if has_drug_combos:
        raise DrugCombosNotImplementedError()
    else:
        # TODO: Support drug combos
        expt_data = expt_data.reset_index(['drug', 'dose'])
        expt_data['drug'] = expt_data['drug'].apply(lambda x: x[0])
        expt_data['dose'] = expt_data['dose'].apply(lambda x: x[0])
        expt_data.set_index(['drug', 'dose'], append=True,
                            inplace=True)
        drugs = expt_data.index.get_level_values('drug').unique()

    if len(drugs) > 1 and len(cell_lines) == 1:
        group_by = ['drug']
    elif len(cell_lines) > 1 and len(drugs) == 1:
        group_by = ['cell_line']
    else:
        group_by = ['cell_line', 'drug']

    if datasets is not None and len(datasets) > 1:
        if len(cell_lines) == 1 and len(drugs) == 1:
            group_by = ['dataset']
        else:
            group_by = ['dataset'] + group_by

    fit_params = []

    is_viability = 'viability' in expt_data.columns

    for group_name, dip_grp in expt_data.groupby(level=group_by):
        if 'dataset' in group_by:
            if len(group_by) > 1:
                dataset = group_name[group_by.index('dataset')]
            else:
                dataset = group_name
        elif datasets is not None:
            dataset = datasets[0]
        else:
            dataset = None

        if 'cell_line' in group_by:
            if len(group_by) > 1:
                cl_name = group_name[group_by.index('cell_line')]
            else:
                cl_name = group_name
        else:
            cl_name = cell_lines[0]

        if 'drug' in group_by:
            if len(group_by) > 1:
                dr_name = group_name[group_by.index('drug')]
            else:
                dr_name = group_name
        else:
            dr_name = drugs[0]

        doses_expt = dip_grp.index.get_level_values('dose').values

        if is_viability:
            resp_expt = dip_grp['viability']
            doses = doses_expt
            fit_obj = fit_drc(
                doses_expt, resp_expt, response_std_errs=None,
                null_rejection_threshold=None,
                fit_cls=fit_cls
            )
            aa_obs_val = aa_obs(resp_expt, doses_expt)
        else:
            if dataset is None and ctrl_data is not None and \
                    'dataset' in ctrl_data.index.names:
                raise ValueError('Experimental data does not have "dataset" '
                                 'in index, but control data does. Please '
                                 'make sure "dataset" is in both dataframes, '
                                 'or neither.')

            ctrl_dip_data_cl = \
                _get_control_responses(ctrl_data, dataset, cl_name,
                                       dip_grp)
            dip_ctrl = []
            dip_ctrl_std_err = []
            if ctrl_dip_data_cl is not None:
                dip_ctrl = ctrl_dip_data_cl['dip_rate'].values
                dip_ctrl_std_err = ctrl_dip_data_cl[
                    'dip_fit_std_err'].values

            n_controls = len(dip_ctrl)
            ctrl_dose_val = ctrl_dose_fn(doses_expt)
            doses_ctrl = np.repeat(ctrl_dose_val, n_controls)
            doses = np.concatenate((doses_ctrl, doses_expt))
            resp_expt = dip_grp['dip_rate'].values
            dip_all = np.concatenate((dip_ctrl, resp_expt))
            dip_std_errs = np.concatenate((
                dip_ctrl_std_err,
                dip_grp['dip_fit_std_err'].values))

            try:
                fit_obj = fit_drc(
                    doses, dip_all, dip_std_errs,
                    fit_cls=fit_cls,
                    ctrl_dose_test=True
                )
            except KeyError:
                fit_obj = None

            if fit_obj is not None:
                aa_obs_val = aa_obs(resp_expt / fit_obj.divisor, doses_expt)
            elif len(dip_ctrl) > 0:
                # If no fit, use average ctrl response to put DIP values
                # on relative scale
                aa_obs_val = aa_obs(resp_expt / np.mean(dip_ctrl), doses_expt)
            else:
                aa_obs_val = None

        max_dose_measured = np.max(doses)
        min_dose_measured = np.min(doses)

        fit_data = dict(
            dataset_id=dataset if dataset is not None else '',
            cell_line=cl_name,
            drug=dr_name,
            fit_obj=fit_obj,
            min_dose_measured=min_dose_measured,
            max_dose_measured=max_dose_measured,
            emax_obs=np.min(resp_expt),
            aa_obs=aa_obs_val
        )

        fit_params.append(fit_data)

    df_params = pd.DataFrame(fit_params)
    df_params.set_index(['dataset_id', 'cell_line', 'drug'], inplace=True)

    df_params._drmetric = 'viability' if is_viability else 'dip'
    if is_viability:
        df_params._viability_time = expt_data_orig._viability_time
        df_params._viability_assay = expt_data_orig._viability_assay

    return df_params


def _calc_ic(row, ic_num):
    if row.fit_obj is None:
        return None

    ic_n = row.fit_obj.ic(ic_num=ic_num)
    if ic_n is None:
        return None

    ic_n = np.min((ic_n, row.max_dose_measured))
    ic_n = np.max((ic_n, row.min_dose_measured))
    return ic_n


def _calc_ec(row, ec_num):
    if row.fit_obj is None:
        return None

    ec_n = row.fit_obj.ec(ec_num=ec_num)
    if ec_n is None:
        return None

    ec_n = np.min((ec_n, row.max_dose_measured))
    ec_n = np.max((ec_n, row.min_dose_measured))
    return ec_n


def _calc_e(row, ec_lbl, relative=False):
    if row.fit_obj is None:
        return None

    ec_val = row[ec_lbl]
    if ec_val is None:
        return None

    return row.fit_obj.fit_rel(ec_val) if relative else row.fit_obj.fit(ec_val)


def _calc_aa(row):
    if not row.fit_obj:
        return None

    return row.fit_obj.aa(min_conc=row.min_dose_measured,
                          max_conc=row.max_dose_measured)


def _calc_auc(row):
    if not row.fit_obj:
        return None

    return row.fit_obj.auc(min_conc=row.min_dose_measured)


def _calc_ec50(row):
    if not row.fit_obj or row.fit_obj.ec50 is None:
        return None

    ec50 = np.min((row.fit_obj.ec50, row.max_dose_measured))
    ec50 = np.max((ec50, row.min_dose_measured))
    return ec50


def _attach_extra_params(base_params,
                         custom_ic_concentrations=frozenset(),
                         custom_ec_concentrations=frozenset(),
                         custom_e_values=frozenset(),
                         custom_e_rel_values=frozenset(),
                         include_aa=False,
                         include_auc=False,
                         include_hill=False,
                         include_emax=False,
                         include_einf=False
                         ):
    datasets = base_params.index.get_level_values('dataset_id').unique()
    if len(datasets) == 1 and datasets[0] == '':
        datasets = None

    cell_lines = base_params.index.get_level_values('cell_line').unique()
    drugs = base_params.index.get_level_values('drug').unique()

    if len(drugs) > 1 and len(cell_lines) == 1:
        group_by = ['drug']
    elif len(cell_lines) > 1 and len(drugs) == 1:
        group_by = ['cell_line']
    else:
        group_by = ['cell_line', 'drug']

    if datasets is not None and len(datasets) > 1:
        if len(cell_lines) == 1 and len(drugs) == 1:
            group_by = ['dataset_id']
        else:
            group_by = ['dataset_id'] + group_by

    index_names = base_params.index.names

    def _generate_label(index):
        group_name_components = []

        if 'dataset_id' in group_by:
            dataset = index[index_names.index('dataset_id')]
            group_name_components.append(str(dataset))

        if 'cell_line' in group_by:
            cl_name = index[index_names.index('cell_line')]
            group_name_components.append(str(cl_name))

        if 'drug' in group_by:
            dr_name = index[index_names.index('drug')]
            group_name_components.append(str(dr_name))

        return "\n".join(group_name_components)

    base_params['label'] = base_params.index.map(_generate_label)

    for ic_num in custom_ic_concentrations:
        base_params['ic{:d}'.format(ic_num)] = base_params.apply(
            _calc_ic, args=(ic_num,), axis=1)

    if 50 in custom_ec_concentrations:
        base_params['ec50'] = base_params.apply(_calc_ec50, axis=1)

    if include_emax:
        base_params['emax'] = base_params.apply(
            lambda row: None if not row.fit_obj else row.fit_obj.fit(
                row.max_dose_measured),
            axis=1)

    if include_einf:
        base_params['einf'] = base_params.apply(
            lambda row: None if not row.fit_obj else row.fit_obj.emax,
            axis=1
        )

    is_viability = base_params._drmetric == 'viability'

    if not is_viability and include_emax:
        divisor = base_params['fit_obj'].apply(lambda fo: fo.divisor if fo
            else None)
        base_params['emax_rel'] = base_params['emax'] / divisor
        base_params['emax_obs_rel'] = base_params['emax_obs'] / divisor

    if include_aa:
        base_params['aa'] = base_params.apply(_calc_aa, axis=1)

    if include_auc:
        base_params['auc'] = base_params.apply(_calc_auc, axis=1)

    if include_hill:
        base_params['hill'] = base_params['fit_obj'].apply(
            lambda fo: fo.hill_slope if fo else None)

    custom_ec_concentrations = set(custom_ec_concentrations)
    custom_ec_concentrations.discard(50)
    custom_ec_concentrations = custom_ec_concentrations.union(
        custom_e_values)
    custom_ec_concentrations = custom_ec_concentrations.union(
        custom_e_rel_values)

    for ec_num in custom_ec_concentrations:
        base_params['ec{:d}'.format(ec_num)] = base_params.apply(
            _calc_ec, args=(ec_num,), axis=1)

    for e_num in custom_e_values:
        base_params['e{:d}'.format(e_num)] = base_params.apply(
            _calc_e, args=('ec{:d}'.format(e_num), False), axis=1
        )

    for e_num in custom_e_rel_values:
        base_params['e{:d}_rel'.format(e_num)] = base_params.apply(
            _calc_e, args=('ec{:d}'.format(e_num), True), axis=1
        )

    return base_params


def _attach_response_values(df_params, ctrl_dip_data, expt_dip_data,
                            ctrl_dose_fn):
    is_viability = df_params._drmetric == 'viability'
    data_list = []
    for grp, dip_grp in expt_dip_data.groupby(
            ['dataset', 'cell_line', 'drug'], sort=False):
        # Assumes drug combinations have been ruled out by fit_params_minimal
        doses_expt = [d[0] for d in dip_grp.index.get_level_values(
            'dose').values]
        fit_data = {'dataset_id': grp[0], 'cell_line': grp[1], 'drug': grp[
            2][0]}

        ctrl_dip_data_cl = \
            _get_control_responses(ctrl_dip_data, grp[0], grp[1],
                                   dip_grp)
        if ctrl_dip_data_cl is not None:
            if is_viability:
                ctrl_dip_data_cl = ctrl_dip_data_cl.to_frame()
            n_controls = len(ctrl_dip_data_cl.index)
            ctrl_dose_val = ctrl_dose_fn(doses_expt)
            doses_ctrl = np.repeat(ctrl_dose_val, n_controls)
            ctrl_dip_data_cl['dose'] = doses_ctrl
            ctrl_dip_data_cl.reset_index('well_id', inplace=True)
            ctrl_dip_data_cl.set_index(['dose', 'well_id'], inplace=True)
            if is_viability:
                fit_data['viability_ctrl'] = ctrl_dip_data_cl['value']
            else:
                fit_data['dip_ctrl'] = ctrl_dip_data_cl['dip_rate']

        if is_viability:
            fit_data['viability_time'] = dip_grp['timepoint'].values
            fit_data['viability'] = pd.Series(
                data=dip_grp['viability'].values,
                index=[doses_expt, dip_grp.index.get_level_values(
                    'well_id')]
            )
            fit_data['viability'].index.rename(['dose', 'well_id'],
                                               inplace=True)
        else:
            fit_data['dip_expt'] = pd.Series(
                data=dip_grp['dip_rate'].values,
                index=[doses_expt,
                       dip_grp.index.get_level_values('well_id')]
            )
            fit_data['dip_expt'].index.rename(['dose', 'well_id'],
                                              inplace=True)

        data_list.append(fit_data)

    df = pd.DataFrame(data_list)
    df.set_index(['dataset_id', 'cell_line', 'drug'], inplace=True)
    df_params_old = df_params
    df_params = pd.concat([df, df_params], axis=1)

    df_params._drmetric = df_params_old._drmetric
    if is_viability:
        df_params._viability_time = df_params_old._viability_time
        df_params._viability_assay = df_params_old._viability_assay

    return df_params


def fit_params(ctrl_data, expt_data,
               fit_cls=HillCurveLL4,
               ctrl_dose_fn=lambda doses: np.min(doses) / CTRL_DOSE_DIVISOR):
    """
    Fit dose response curves to DIP rates or viability data

    This method computes parameters including IC50, EC50, AUC, AA,
    Hill coefficient, and Emax. For a faster version,
    see :func:`fit_params_minimal`.

    Parameters
    ----------
    ctrl_data: pd.DataFrame or None
        Control DIP rates from :func:`dip_rates` or :func:`ctrl_dip_rates`.
        Set to None to not use control data.
    expt_data: pd.DataFrame
        Experiment (non-control) DIP rates from :func:`dip_rates` or
        :func:`expt_dip_rates`, or viability data from :func:`viability`
    fit_cls: Class
        Class to use for curve fitting (default: :func:`HillCurveLL4`)
    ctrl_dose_fn: function
        Function to use to set an effective "dose" (non-zero) for controls.
        Takes the list of experiment doses as an argument.

    Returns
    -------
    pd.DataFrame
        DataFrame containing DIP rate curve fits and parameters
    """
    base_params = fit_params_minimal(ctrl_data, expt_data, fit_cls,
                                     ctrl_dose_fn)

    return fit_params_from_base(
        base_params, ctrl_data, expt_data,
        ctrl_dose_fn=ctrl_dose_fn,
        custom_ic_concentrations={50},
        custom_ec_concentrations={50},
        include_auc=True,
        include_aa=True,
        include_hill=True,
        include_emax=True,
        include_einf=True,
        include_response_values=True
    )


def fit_params_from_base(
        base_params,
        ctrl_resp_data=None, expt_resp_data=None,
        ctrl_dose_fn=lambda doses: np.min(doses) / CTRL_DOSE_DIVISOR,
        custom_ic_concentrations=frozenset(),
        custom_ec_concentrations=frozenset(),
        custom_e_values=frozenset(),
        custom_e_rel_values=frozenset(),
        include_aa=False,
        include_auc=False,
        include_hill=False,
        include_emax=False,
        include_einf=False,
        include_response_values=True):
    """
    Attach additional parameters to basic set of fit parameters
    """
    df_params = _attach_extra_params(base_params, custom_ic_concentrations,
                                     custom_ec_concentrations,
                                     custom_e_values, custom_e_rel_values,
                                     include_aa,
                                     include_auc,
                                     include_hill,
                                     include_emax, include_einf)

    if include_response_values:
        df_params = _attach_response_values(df_params, ctrl_resp_data,
                                            expt_resp_data, ctrl_dose_fn)

    return df_params


def is_param_truncated(df_params, param_name):
    """
    Checks if parameter values are truncated at boundaries of measured range

    Parameters
    ----------
    df_params: pd.DataFrame
        DataFrame of DIP curve fits with parameters from :func:`fit_params`
    param_name: str
        Name of a parameter, e.g. 'ic50'

    Returns
    -------
    np.ndarray
        Array of booleans showing whether each entry in the DataFrame is
        truncated
    """
    values = df_params[param_name].fillna(value=np.nan)
    return np.isclose(
        values, df_params['max_dose_measured'],
        atol=PARAM_EQUAL_ATOL, rtol=PARAM_EQUAL_RTOL) | np.isclose(
        values, df_params['min_dose_measured'],
        atol=PARAM_EQUAL_ATOL, rtol=PARAM_EQUAL_RTOL
    )
